/*
** Module   : AgentPool.java
** Abstract : Definition and implementation of agent pools (based on appserver's operating mode).
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20130529 Created initial version.
** 002 CA  20130628 Fixed context-local data reset for State-reset operating mode.
** 003 CA  20130704 Reworked appserver launching - implemented by the AppServerLauncher class.
**                  Appserver definition was moved to its own class.
** 004 CA  20130823 Changes related to support async requests.
*/

package com.goldencode.p2j.util;

import java.util.*;

import org.apache.commons.logging.*;
import com.goldencode.p2j.*;
import com.goldencode.p2j.main.*;
import com.goldencode.p2j.util.Agent.InvocationModes;

/**
 * The main class for agent pooling.  Based on the appserver's operating mode, the 
 * {@link #newConnection()} call will return a:
 * <ul>
 * <li> {@link BoundPool} for State-reset and State-aware modes.</li>
 * <li> {@link UnboundPool} for State-free and Stateless modes.</li>
 * </ul>
 * Agents are extracted from {@link #mainPool} as connection or procedure requests are incoming
 * and put back in the pool only after the connection terminates or it becomes unbound.
 */
abstract class AgentPool
{
   /** Logger. */
   private static final Log LOG = LogFactory.getLog(AgentPool.class);

   /** Map of connection IDs to appserver name. */
   private static final Map<String, String> idsToAppserver = new HashMap<String, String>();

   /** The next connection ID; starts from a random value. */
   private static long nextId = new Random().nextLong();

   /** The pool of agents.*/
   private final Set<Agent> mainPool = new HashSet<Agent>();
   
   /** All started agents. */
   private final Set<Agent> all = new HashSet<Agent>();
   
   /** The remote persistent procedures exposed as proxy procedures to the requester. */
   private final Map<String, Map<String, PersistentProcedure>>
      persistentProcedures = new HashMap<String, Map<String, PersistentProcedure>>();

   /** The appserver definition. */
   private final AppServerDefinition appDef;
   
   /** The EXPORTS for this appserver. */
   private String exports = null;
   
   /**
    * Build a new agent pool, for the specified appserver.
    * 
    * @param    def
    *           The appserver definition.
    */
   private AgentPool(AppServerDefinition def)
   {
      this.appDef = def;
   }
   
   /**
    * Get the appserver name used to established the given connection.
    * 
    * @param    connectionId
    *           The connection ID.
    *           
    * @return   See above.
    */
   static String getAppserver(String connectionId)
   {
      synchronized (idsToAppserver)
      {
         return idsToAppserver.get(connectionId);
      }
   }
   
   /**
    * Instantiate a new agent pool for the given appserver, based on the operating mode.
    * 
    * @param    appServer
    *           The appserver's name.
    *           
    * @return   A new pool instance.
    * 
    * @throws   IllegalArgumentException
    *           If the operating mode can not be resolved to one of {@link AppServerOperatingMode}
    *           constants.
    */
   static AgentPool newPool(String appServer)
   {
      AppServerDefinition appDef = AppServerDefinition.get(appServer);
      switch (appDef.getOperatingMode())
      {
         case STATE_AWARE:
            return new BoundPool(appDef, false);
         case STATE_RESET:
            return new BoundPool(appDef, true);
         case STATELESS:
            return new UnboundPool(appDef, false);
         case STATE_FREE:
            return new UnboundPool(appDef, true);
      }
   
      throw new IllegalArgumentException(
         "Illegal operating mode: " + appDef.getOperatingMode() + ".");
   }

   /**
    * Get the next connection ID.
    * 
    * @return   See above.
    */
   protected static synchronized String nextId()
   {
      do
      {
         String next = Long.toString(++nextId);
         synchronized (idsToAppserver)
         {
            if (!idsToAppserver.containsKey(next))
               return next;
         }
      }
      while (true);
   }

   /**
    * Acquire a new agent from the pool to handle requests incoming via the given connection ID.
    * 
    * @param    id
    *           The connection ID.
    * @param    code
    *           The code of a persistent external procedure to which the agent is associated. May
    *           be <code>null</code>.
    *           
    * @return   See above.
    */
   abstract Agent acquireAgent(String id, String code);

   /**
    * Release the given agent and place it back in the pool. This will handle any bound-related
    * constraints.
    * 
    * @param    id
    *           The connection ID.
    * @param    agent
    *           The agent which needs to be released.
    */
   abstract void releaseAgent(String id, Agent agent);

   /**
    * Establish a new connection.
    * 
    * @return   The connection ID.
    * 
    * @throws   NumberedException
    *           If the connection could not be established.
    */
   abstract String newConnection()
   throws NumberedException;

   /**
    * Get the exports for this appserver.  If set, only the external procedures which match these
    * exports will be available for invocation.
    * 
    * @return   See above.
    */
   synchronized character getExports()
   {
      return new character(exports);
   }
   
   /**
    * Set the exports for this appserver.
    * 
    * @param    exports
    *           The appserver's exports. 
    */
   void setExports(final character exports)
   {
      this.exports = (exports.isUnknown() ? null : exports.toStringMessage());
   }
   
   /**
    * Get the name of the appserver handled by this pool.
    * 
    * @return   See above.
    */
   String getName()
   {
      return appDef.getAppServer();
   }

   /**
    * Get the startup procedure for the appserver handled by this pool.
    * 
    * @return   See above.
    */
   String getStartup()
   {
      return appDef.getStartup();
   }

   /**
    * Get the startup parameters for the appserver handled by this pool.
    * 
    * @return   See above.
    */
   String getStartupParameter()
   {
      return appDef.getStartupParameter();
   }

   /**
    * Get the shutdown procedure for the appserver handled by this pool.
    * 
    * @return   See above.
    */
   String getShutdown()
   {
      return appDef.getShutdown();
   }

   /**
    * Get the activate procedure for the appserver handled by this pool.
    * 
    * @return   See above.
    */
   String getActivate()
   {
      return appDef.getActivate();
   }

   /**
    * Get the deactivate procedure for the appserver handled by this pool.
    * 
    * @return   See above.
    */
   String getDeactivate()
   {
      return appDef.getDeactivate();
   }

   /**
    * Establish a new appserver connection. This will invoke any connect procedure for this 
    * appserver, to which the given <code>user</code>, <code>pwd</code> and <code>serverInfo</code>
    * parameters will be passed.
    * <p>
    * The appserver connection must be configured to start under the given P2J-level subject ID.
    *  
    * @param    subjectId
    *           The subject ID associated with this appserver.
    * @param    sessionFree
    *           Flag indicating if the connection request is for a session-free operating mode.
    * @param    user
    *           The first parameter to pass to the connect procedure, if any.
    * @param    pwd
    *           The second parameter to pass to the connect procedure, if any.
    * @param    serverInfo
    *           The third parameter to pass to the connect procedure, if any.
    *           
    * @return   An {@link AppServerInvocationResult} instance with any result or caught errors.
    */
   AppServerInvocationResult connect(String  subjectId,
                                     boolean sessionFree,
                                     String  user,
                                     String  pwd,
                                     String  serverInfo)
   {
      AppServerInvocationResult result = new AppServerInvocationResult();
   
      AppServerOperatingMode operatingMode = getOperatingMode();
      if (operatingMode.isSessionFree() != sessionFree)
      {
         final String msg =
            "The %s connection failed because the AppServer is in a %s operating mode (%s)";
         final String err = String.format(msg,
                                          sessionFree ? "Session-free" : "Session-managed",
                                          sessionFree ? "session-managed" : "session-free",
                                          operatingMode);
         result.setError(new NumberedException(err, 13511));
         return result;
      }
   
      String id = null;
      try
      {
         id = newConnection();
      }
      catch (NumberedException e)
      {
         result.setError(e);
         return result;
      }
      
      synchronized(idsToAppserver)
      {
         idsToAppserver.put(id, appDef.getAppServer());
      }
      
      String connect = appDef.getConnect();
      if (!operatingMode.isSessionFree() && connect != null && connect.length() > 0)
      {
         Agent agent = acquireAgent(id, null);
         agent.invoke(result, id, Agent.InvocationModes.AGENT_WAIT, connect, null, "Connect", 
                      "III", new character(user), new character(pwd), new character(serverInfo));
         releaseAgent(id, agent);
         if (result.getError() != null)
         {
            // in case of error, disconnect
            disconnect(id);
   
            id = null;
         }
      }
      
      result.setConnectionID(id);
      
      // regardless of OK/error, the result is sent
      return result;
   }

   /**
    * Disconnect the appserver connection, with the given ID.  It will invoke any configured 
    * disconnect procedure, but will not return any errors or values to the caller.
    * 
    * @param    id
    *           The connection ID to disconnect.
    */
   void disconnect(String id)
   {
      AppServerOperatingMode operatingMode = getOperatingMode();
      
      String disconnect = appDef.getDisconnect();
      if (!operatingMode.isSessionFree() && disconnect != null && disconnect.length() > 0)
      {
         boolean stateless = (appDef.getOperatingMode() == AppServerOperatingMode.STATELESS);
         AppServerInvocationResult result = new AppServerInvocationResult();
         
         Agent agent = acquireAgent(id, null);
         // call the disconnect procedure.
         InvocationModes mode = (stateless ? InvocationModes.AGENT_WAIT 
                                           : InvocationModes.AGENT_NO_WAIT);
         agent.invoke(result, id, mode, disconnect, null, "Disconnect", null);
      }
      else
      {
         terminateConnection(id);
      }
   }

   /**
    * Start listening for appserver-related commands.
    */
   void start()
   {
      // this is a new Agent for the appServer. do not allow to start more than maxAgents
      int max = appDef.getMaxAgents();
      if (size() >= max)
      {
         if (LOG.isWarnEnabled())
         {
            LOG.warn("Pool capacity of " + max + " reached, no more agents will be started.");
         }

         // just terminate the client, nothing else to do
         return;
      }
      
      // add it to the pool and let it listen for commands
      Agent agent = new Agent(this, appDef.getPropath());
      poolAgent(agent, true);
      try
      {
         agent.listen();
      }
      finally
      {
         // the agent was terminated, remove it from the pool
         removeAgent(agent);
      }
   }

   /**
    * Send the terminate command to all agents in this pool.
    */
   void terminateAll()
   {
      for (Agent agent : all)
      {
         agent.terminate();
      }
   }

   /**
    * Terminate the connection with the given ID.
    * 
    * @param    id
    *           The connection ID.
    */
   void terminateConnection(String id)
   {
      synchronized(idsToAppserver)
      {
         idsToAppserver.remove(id);
      }
   
      Map<String, PersistentProcedure> procs = null;
      synchronized (persistentProcedures)
      {
         procs = persistentProcedures.get(id);
   
         if (procs != null)
         {
            // delete all resources, but use a copy as iterator
            procs = new HashMap<String, PersistentProcedure>(procs);
            for (PersistentProcedure proc : procs.values())
            {
               removeProcedure(id, new handle(proc).toString());
               
               proc.delete();
            }
         }
         
         persistentProcedures.remove(id);
      }
   }

   /**
    * Get the operating mode for this appserver.
    * 
    * @return   See above.
    */
   AppServerOperatingMode getOperatingMode()
   {
      return appDef.getOperatingMode();
   }

   /**
    * Add an agent back to the {@link #mainPool}.
    * 
    * @param    agent
    *           The agent to add back to the pool.
    * @param    newlyCreated
    *           When this flag is <code>true</code>, the agent is added to the {@link #all global}
    *           pool too.
    */
   void poolAgent(Agent agent, boolean newlyCreated)
   {
      agent.unbind();
      
      synchronized (mainPool)
      {
         mainPool.add(agent);
         mainPool.notify();
      }
   
      if (newlyCreated)
      {
         synchronized (all)
         {
            all.add(agent);
            all.notify();
         }
      }
   }

   /**
    * Check if this pool is empty.
    * 
    * @return   <code>true</code> if the pool is empty, <code>false</code> otherwise.
    */
   boolean isEmpty()
   {
      synchronized (all)
      {
         return all.isEmpty();
      }
   }
   
   /**
    * Remove the given agent entirely from the pool.
    * 
    * @param    agent
    *           The agent which was terminated and needs to be removed.
    */
   void removeAgent(Agent agent)
   {
      synchronized (all)
      {
         all.remove(agent);
         all.notify();
      }
      
      synchronized (mainPool)
      {
         mainPool.remove(agent);
      }
   }

   /**
    * Get the pool size.
    * 
    * @return   See above.
    */
   int size()
   {
      synchronized (all)
      {
         return all.size();
      }
   }

   /**
    * Save the given persistent procedure and associate it with the given connection ID.
    * 
    * @param    id
    *           The connection ID.
    * @param    procId
    *           The procedure's ID.
    * @param    proc
    *           The persistent procedure.
    */
   void saveProcedure(String id, String procId, PersistentProcedure proc)
   {
      Map<String, PersistentProcedure> procs = null;
      synchronized (persistentProcedures)
      {
         procs = persistentProcedures.get(id);
         if (procs == null)
         {
            persistentProcedures.put(id, procs = new HashMap<String, PersistentProcedure>());
         }
      }
      
      procs.put(procId, proc);
   }

   /**
    * Check if the given persistent procedure was started as a remote procedure with the appserver
    * associated with this pool.
    * 
    * @param    proc
    *           The persistent procedure.
    *           
    * @return   See above.
    */
   boolean hasProcedure(PersistentProcedure proc)
   {
      synchronized (persistentProcedures)
      {
         for (Map<String, PersistentProcedure> procs : persistentProcedures.values())
         {
            if (procs.containsValue(proc))
               return true;
         }
      }
      
      return false;
   }
   
   /**
    * Remove the specified persistent procedure.
    * 
    * @param    id
    *           The connection ID.
    * @param    procId
    *           The procedure ID.
    *           
    * @return   The procedure with the given ID or null if non exists.
    */
   PersistentProcedure removeProcedure(String id, String procId)
   {
      synchronized (persistentProcedures)
      {
         return persistentProcedures.get(id).remove(procId);
      }
   }
   
   /**
    * Get the persistent procedure associated with the given ID; this returns a non-null instance
    * only if the procedure was not deleted from the context specified by the connection ID.
    *   
    * @param    id
    *           The connection ID.
    * @param    procId
    *           The procedure ID.
    *           
    * @return   The persistent procedure or <code>null</code> if none found.
    */
   PersistentProcedure getProcedure(String id, String procId)
   {
      synchronized (persistentProcedures)
      {
         return persistentProcedures.get(id).get(procId);
      }
   }

   /**
    * Trim the running agents so the pool size is as close to 
    * {@link AppServerDefinition#getMinAgents()} as possible.
    */
   void trimAgents()
   {
      final int size = appDef.getMinAgents();
      // use a copy, as 
      Set<Agent> copy = null;
      synchronized (all)
      {
         copy = new HashSet<Agent>(all);
      }

      int n = copy.size() - size;
      // check if we can to terminate any agents
      if (n <= 0)
      {
         return;
      }

      for (Agent a : copy)
      {
         synchronized (mainPool)
         {
            // if this agent is not busy, terminate it
            if (mainPool.contains(a))
            {
               // remove it from main pool first, so it can't be picked by other connections
               mainPool.remove(a);
               
               // terminate
               a.terminate();
            }
         }

         // exit if no more trimming possible
         if (--n == 0)
         {
            break;
         }
      }
   }

   /**
    * Extract an agent from the main pool; if non available, attempt to start one. If the max
    * number of running agents is reached and all agents are busy, return <code>null</code>.
    * 
    * @return   The extracted agent or <code>null</code> if all agents are busy.
    */
   Agent getFromMainPool()
   {
      Agent agent = null;
      long timeout = appDef.getRequestTimeout() * 1000;
      synchronized (mainPool)
      {
         if (mainPool.size() == 0)
         {
            try
            {
               mainPool.wait(timeout);
            }
            catch (InterruptedException e)
            {
               // ignore
            }
         }
   
         if (mainPool.size() > 0)
         {
            agent = mainPool.iterator().next();
            mainPool.remove(agent);
         }
      }

      int prevSize = size();
      if (agent == null && prevSize < appDef.getMaxAgents())
      {
         // automatically start an agent, until maxAgents is reached
         AppServerLauncher.launch(appDef.getAppServer());
         
         // if an agent was started, get it from the main pool.
         if (size() > prevSize)
         {
            agent = getFromMainPool();
         }
      }
      
      return agent;
   }

   /**
    * Class implementing the behavior for STATE-RESET and STATE-AWARE operating modes.  In these
    * modes, an agent is bound to the connection when the connection is established.
    */
   static class BoundPool
   extends AgentPool
   {
      /** The reserved agent, for each connection. */
      private final Map<String, Agent> connections = new HashMap<String, Agent>();

      /** Flag indicating this is STATE-RESET mode. */
      private final boolean reset;
      
      /**
       * Build a new agent pool, for the specified appserver.
       * 
       * @param    appDef
       *           The appserver definition.
       * @param    reset
       *           Flag indicating this is a STATE-RESET mode.
       */
      BoundPool(AppServerDefinition appDef, boolean reset)
      {
         super(appDef);
         this.reset = reset;
      }
      
      /**
       * Get the agent associated with the given connection id.
       * 
       * @param    id
       *           The connection ID.
       * @param    code
       *           The code of a persistent external procedure to which the agent is associated. 
       *           Not used for bound pools (as async requests are not possible).
       *           
       * @return   See above.
       */
      Agent acquireAgent(String id, String code)
      {
         synchronized (connections)
         {
            return connections.get(id);
         }
      }
      
      /**
       * This is a no-op, as the agent is released only when the connection is terminated.
       * 
       * @param    id
       *           The connection ID.
       * @param    agent
       *           The agent which needs to be released.
       */
      void releaseAgent(String id, Agent agent)
      {
         // no-op, agent is kept bound for the duration of the connection
      }
      
      /**
       * Establish a new connection.
       * 
       * @return   The connection ID.
       * 
       * @throws   NumberedException
       *           If the connection could not be established.
       */
      String newConnection()
      throws NumberedException
      {
         String id = nextId();
         // must reserve one agent for this bound connection
         Agent agent = getFromMainPool();
         if (agent == null)
         {
            throw new NumberedException("No servers available");
         }
         
         synchronized (connections)
         {
            connections.put(id, agent);
         }

         agent.bind();
         
         return id;
      }

      /**
       * Terminate the connection with the given ID and unbind the agent.
       * 
       * @param    id
       *           The connection ID.
       */
      void terminateConnection(String id)
      {
         super.terminateConnection(id);
         
         Agent agent = null;
         synchronized (connections)
         {
            agent = connections.remove(id);
         }
         
         if (reset)
         {
            // reset the agent's context
            agent.resetContext();
         }

         poolAgent(agent, false);
      }
   }
   
   /**
    * Class implementing the behavior for STATE-FREE and STATELESS operating modes.  In these
    * modes, no agent is bound to the connection when the connection is established.
    */
   static class UnboundPool
   extends AgentPool
   {
      /**
       * The reserved agent, for each connection.  This map is populated only when the agent
       * becomes bound.
       */
      private final Map<String, Agent> connections = new HashMap<String, Agent>();
      
      /** Flag indicating this is a STATE-FREE mode. */
      private final boolean free;
      
      /** Map of procedure codes to their associated agent. Used in async requests. */
      private final Map<String, Agent> procedures = new HashMap<String, Agent>();
      
      /**
       * Build a new agent pool, for the specified appserver.
       * 
       * @param    appDef
       *           The appserver definition.
       * @param    free
       *           Flag indicating this is a STATE-FREE mode.
       */
      UnboundPool(AppServerDefinition appDef, boolean free)
      {
         super(appDef);
         this.free = free;
      }

      /**
       * Get the agent associated with the given connection id.  If no agent is yet bound, it
       * extracts a new agent from the pool.
       * 
       * @param    id
       *           The connection ID.
       * @param    code
       *           The code of a persistent external procedure to which the agent is associated.
       *           May be <code>null</code>.
       *           
       * @return   See above.
       */
      Agent acquireAgent(String id, String code)
      {
         Agent agent = null;
         
         // first check if we have a bound agent
         synchronized (connections)
         {
            agent = connections.get(id);
         }
         
         if (agent == null && code != null)
         {
            // no connection-bound agent, get the dedicated agent, if a procedure code is specified
            agent = getByCode(code);
         }
         
         if (agent == null)
         {
            // last resort, get one from main pool
            agent = getFromMainPool();
         }
         
         return agent;
      }

      /**
       * Release the given agent and place it back in the pool, but only if is still unbound.
       * 
       * @param    id
       *           The connection ID.
       * @param    agent
       *           The agent which needs to be released.
       */
      void releaseAgent(String id, Agent agent)
      {
         String code = agent.getProcedure() == null ? null 
                                                    : new handle(agent.getProcedure()).toString();

         if (agent.isBound())
         {
            // agent became bound, save it
            if (code != null)
            {
               synchronized (procedures)
               {
                  procedures.put(code, agent);
               }
            }
            else if (!free)
            {
               // in non-free mode, agents are bound only when a procedure is invoked PERSISTENT
               synchronized (connections)
               {
                  connections.put(id, agent);
               }
            }
         }
         else
         {
            if (!free)
            {
               synchronized (connections)
               {
                  connections.remove(id);
               }
            }
            
            if (code != null)
            {
               synchronized (procedures)
               {
                  procedures.remove(code);
               }

               agent.setProcedure(null);
            }
            
            // if there was a bound agent, add it back to the main pool
            poolAgent(agent, false);
         }
      }
      
      /**
       * Establish a new connection.
       * 
       * @return   The connection ID.
       * 
       * @throws   NumberedException
       *           If the connection could not be established.
       */
      String newConnection()
      {
         String id = nextId();
      
         synchronized (connections)
         {
            // no agent is initially bound to this pool
            connections.put(id, null);
         }
      
         return id;
      }

      /**
       * Terminate the connection with the given ID and unbind the agent.
       * 
       * @param    id
       *           The connection ID.
       */
      void terminateConnection(String id)
      {
         super.terminateConnection(id);

         Agent agent = null;
         synchronized (connections)
         {
            agent = connections.remove(id);
         }

         if (agent != null)
         {
            // if there was a bound agent, add it back to the main pool
            poolAgent(agent, false);
         }
      }
      
      /**
       * Remove the specified persistent procedure.
       * 
       * @param    id
       *           The connection ID.
       * @param    procId
       *           The procedure ID.
       *           
       * @return   The procedure with the given ID or null if non exists.
       */
      PersistentProcedure removeProcedure(String id, String procId)
      {
         Agent agent = null;
         synchronized (procedures)
         {
            agent = procedures.get(procId);
         }

         if (agent != null)
         {
            // if there was a bound agent, release it
            agent.unbind();
            releaseAgent(id, agent);
            agent.setProcedure(null);
         }
         
         return super.removeProcedure(id, procId);
      }

      /**
       * Get the agent associated with the procedure identified by the given code.
       * 
       * @param    procId
       *           The code of a persistent external procedure to which the agent is associated. 
       *           
       * @return   See above.
       */
      private Agent getByCode(String procId)
      {
         synchronized (procedures)
         {
            return procedures.get(procId);
         }
      }
   }
}
